"""
安全通信远程证明功能.
"""

__all__ = [
    "RaConfig",
    "RaPod",
    "RaRequest",
    "RaResponse",
    "RaResponseMetadata",
    "RaResponsePod",
    "RaResponsePods",
    "RaResponseUrl",
    "attest_server",
    "prepare_ra_request",
    "validata_ra_request",
    "attest_client",
]

import base64
import json
import requests

from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing_extensions import Dict, List, Literal, NotRequired, Optional, TypedDict, Union

from . import error
from .logger_factory import LoggerFactory
from .utils import random_str, request_bytedance_gateway

RA_NONCE_LEN = 12
RA_TYPE_LOCAL = "local"
RA_TYPE_TCA = "tca"


class RaPod(TypedDict):
    cluster_id: str

    namespace: str

    deployment: str

    pod: NotRequired[Literal["random", "all"]]
    """默认为 random, 如果想要全部则为 all."""


@dataclass
class RaConfig:
    """
    访问远程证明功能的配置.
    """

    ra_url: str
    """远程证明端点的 HTTP 地址."""

    ra_type: str = RA_TYPE_TCA
    """ra的类型，local：本地RA， tca：TCA的RAS_adapter接口(默认)."""

    ra_service_name: str = ""
    """数据接收端（服务端）的服务名. ,用于验证通过机密容器部署的服务"""

    ra_pods_info: str = ""
    """
        待验证的节点的pod信息，用于验证非通过机密容器部署的服务, 
        格式: {cluster_id:***, namespace:***, deployment:***}
        ra_service_name 和 ra_pods_info 二选一
    """

    ra_uid: str = ""
    """用户的UID"""

    ra_key_negotiation: bool = True
    """是否在验证的时候需要同步进行协商密钥."""

    ra_need_token: bool = True  # RA 功能稳定后要改成 False
    """是否要返回验证 token. 默认为是."""

    ra_policy_id: str = ""
    """进行验证所使用的策略, 如果 ra_need_token 为 True, 则必填."""

    bytedance_top_info: str = ""
    """访问bytedance TOP网关的信息."""

    ra_attested_pods: List[RaPod] = field(default_factory=list)
    """old 需要验证的节点信息."""

    ra_policy_ids: List[str] = field(default_factory=list)
    """old 进行验证所使用的策略, 如果 ra_need_token 为 True, 则必填."""


class RaRequest(TypedDict):
    """数据发送方发出的远程证明请求."""

    key_negotiation: bool
    """是否在验证的时候需要同步进行协商密钥."""

    token: bool
    """是否要返回验证 token."""

    policy_ids: List[str]
    """进行验证所使用的策略."""

    attested_pods: List[RaPod]
    """需要验证的节点信息."""


class RaResponseMetadata(TypedDict):
    Code: int
    Message: str


class RaResponseUrl(TypedDict):
    result_url: str


class RaResponsePod(TypedDict):
    evidence: str
    token: str
    key_info: dict


RaResponsePods = Dict[str, RaResponsePod]


class RaResponse(TypedDict):
    """数据接收方发出的远程证明响应."""

    ResponseMetadata: RaResponseMetadata

    ResponseResult: Union[RaResponseUrl, RaResponsePods]


def prepare_ra_request(config: RaConfig) -> RaRequest:
    """
    准备远程证明请求.

    Returns:
        远程证明请求. 可作为 JSON 发送给数据接收方远程证明端点.
    """
    return RaRequest(
        key_negotiation=config.ra_key_negotiation,
        token=config.ra_need_token,
        policy_ids=config.ra_policy_ids,
        attested_pods=config.ra_attested_pods,
    )


def validata_ra_request(request: RaRequest) -> None:
    """
    验证远程证明请求有效性.

    Raises:
        ParamError: 请求的 nonce 或 timestamp 字段无效.
    """
    nonce = request.get("hex_runtime_data")
    if nonce:
        nonce = base64.b64decode(nonce)
        if len(nonce) != RA_NONCE_LEN:
            raise error.ParamError("nonce", "Invalid RA request")

    # time = datetime.fromtimestamp(request["timestamp"], timezone.utc)
    # now = datetime.now(timezone.utc)
    # if time > now + timedelta(minutes=1) or time < now - timedelta(minutes=2):
    #     raise error.ParamError("timestamp", "Invalid RA request")


def request_tca(config: RaConfig):
    logger = LoggerFactory.get_local_instance("pcc_jsc_client")
    try:
        ra_url = config.ra_url
        ra_service_name = config.ra_service_name
        ra_pods_info = config.ra_pods_info
        ra_key_negotiation = config.ra_key_negotiation
        ra_policy_id = config.ra_policy_id
        ra_uid = config.ra_uid

        top_info = config.bytedance_top_info
        if top_info and isinstance(top_info, str):
            top_info = json.loads(top_info)
        # assert isinstance(top_info, dict)

        headers = {"UID": ra_uid}

        if ra_service_name:
            body = {
                "PolicyID": ra_policy_id,
                "ServiceName": ra_service_name,
                "KeyNegotiation": ra_key_negotiation,
                "Token": True
            }
        elif ra_pods_info:
            if type(ra_pods_info) == str:
                ra_pods_info = json.loads(ra_pods_info)
            body = {
                "PolicyID": ra_policy_id,
                "AttestedPods": [ra_pods_info],
                "KeyNegotiation": ra_key_negotiation,
                "Token": True
            }
        else:
            raise error.ServiceError("RA", config.ra_url, None, "call TCA error")

        return request_bytedance_gateway(ra_url, body, headers, top_info)
    except Exception as e:
        logger.exception(f"Response is not JSON: service={config.ra_url}")
        raise error.ServiceError("RA", config.ra_url, None, "call TCA error") from e


def attest_server_local(token: Optional[str] = None, config: RaConfig = None) -> tuple:
    """
    对数据接收方进行远程证明(非TCA方式).
    """
    logger = LoggerFactory.get_local_instance("pcc_jsc_client")

    timestamp = str(int(datetime.now(timezone.utc).timestamp()))
    headers = {"timestamp": timestamp}
    if token:
        headers["token"] = f"{token}"

    request = {"nonce": random_str()}
    try:
        logger.info(f"request ra, url={config.ra_url}, header={headers}, params={request}")
        response = requests.post(config.ra_url, headers=headers, json=request)
    except Exception as e:
        logger.exception(f"RA Network error: url={config.ra_url}")
        raise error.NetworkError("RA", config.ra_url, endpoint=None) from e

    if response.status_code != 200:
        logger.error(f"Service error: service={config.ra_url} code={response.status_code}")
        raise error.ServiceError("RA", config.ra_url, None, f"code={response.status_code}")

    try:
        response_json: RaResponse = response.json()
        logger.info(f"ra response={response_json}")

        res_meta_data = response_json.get("ResponseMetadata")
        res_result = response_json.get("ResponseResult")

        if not res_meta_data or not res_result:
            raise error.ServiceError("RA", config.ra_url, None, "wrong format")

        if res_meta_data.get("Code") != 0:
            raise error.ServiceError("RA", config.ra_url, None, res_meta_data.get("Message", ""))

        if result_url := res_result.get("result_url"):
            logger.info(result_url)
            raise NotImplementedError()

        for pod_name, pod_ra_info in res_result.items():
            # evidence = pod_ra_info.get("evidence")
            # token = pod_ra_info.get("token")
            key_info = pod_ra_info.get("key_info")
            if not key_info:
                raise error.ServiceError("RA", config.ra_url, "RSA key", "null")

            return True, key_info["pub_key_info"]
    except Exception as e:
        logger.exception(f"Response is not JSON: service={config.ra_url}")
        raise error.ServiceError("RA", config.ra_url, None, "not JSON") from e


def attest_server(token: Optional[str] = None, config: RaConfig = None) -> tuple:
    """
    对数据接收方进行远程证明.
    """
    logger = LoggerFactory.get_local_instance("pcc_jsc_client")

    if config.bytedance_top_info:
        response = request_tca(config)
    else:
        return attest_server_local(token, config)

    if not response:
        logger.error(f"RA error: url={config.ra_url} response=NULL")
        raise error.ServiceError("RA", config.ra_url, None, "NULL")

    if response.status_code != 200:
        logger.error(f"Service error: service={config.ra_url} code={response.status_code}")
        raise error.ServiceError("RA", config.ra_url, None, f"code={response.status_code}")

    try:
        attestation_status = False
        public_key_info = ""

        ra_res = json.loads(response.text).get("Result")
        if not ra_res:
            logger.error(
                f"attest_server failed, verify_result is NULL, ra interface response={response.text}"
            )
            return False, ""

        for cluster_info, res_info in ra_res.items():
            raw_evidence = res_info.get("evidence")
            tmp = json.loads(raw_evidence)
            raw_evidence_json = json.dumps(tmp, indent=4)
            logger.debug(f"get ra evidence:{raw_evidence_json}")

            token_dict = {}
            raw_token = res_info.get("token")
            token_parts = raw_token.split(".")
            if len(token_parts) >= 2:
                token_data = token_parts[1]
                try:
                    # 使用base64.urlsafe_b64decode进行URL安全模式的解码
                    decoded_data = base64.urlsafe_b64decode(
                        token_data + "=" * (4 - len(token_data) % 4)
                    ).decode("utf-8")

                    token_dict = json.loads(decoded_data)
                except Exception as e:
                    logger.exception(f"解码出现错误: {e}")
                    raise error.ServiceError(
                        "RA", config.ra_url, None, f"code={response.status_code}"
                    )

            # 取第一个节点验证就可以
            attestation_status = len(token_dict.get("policies_matched")) > 0
            public_key_info = res_info.get("key_info").get("pub_key_info")
            break

        return attestation_status, public_key_info
    except Exception as e:
        logger.exception(f"Response is not JSON: service={config.ra_url}")
        raise error.ServiceError("RA", config.ra_url, None, "not JSON") from e


def attest_client(client_ra_config: dict = None, client_bytedance_top_info: str = None) -> bool:
    """
    对数据发送方进行远程证明.
    """
    logger = LoggerFactory.get_local_instance("pcc_jsc_server")

    ra_service_name = client_ra_config.get("ra_service_name")
    ra_policy_id = client_ra_config.get("ra_policy_id")
    ra_pods_info = client_ra_config.get("ra_pods_info")
    ra_key_negotiation = client_ra_config.get("ra_key_negotiation")

    top_info = client_bytedance_top_info
    if top_info and isinstance(top_info, str):
        top_info = json.loads(top_info)

    ra_url = client_ra_config.get("ra_url", "open.volcengineapi.com")
    ra_uid = client_ra_config.get("ra_uid")
    if ra_uid:
        headers = {"UID": ra_uid}
    else:
        headers = {}

    if ra_service_name:
        body = {
            "PolicyID": ra_policy_id,
            "ServiceName": ra_service_name,
            "KeyNegotiation": ra_key_negotiation,
            "Token": True
        }
    elif ra_pods_info:
        if type(ra_pods_info) == str:
            ra_pods_info = json.loads(ra_pods_info)
        body = {
            "PolicyID": ra_policy_id,
            "AttestedPods": [ra_pods_info],
            "KeyNegotiation": ra_key_negotiation,
            "Token": True
        }
    else:
        raise error.ServiceError("RA", ra_service_name, None, "call TCA error")

    response = request_bytedance_gateway(ra_url, body, headers, top_info)
    if not response:
        logger.error(f"Client RA error: response NULL")
        raise error.ServiceError("RA", ra_service_name, None, "NULL")

    if response.status_code != 200:
        logger.error(f"client ra error: ra_service_name={ra_service_name} code={response.status_code}")
        raise error.ServiceError("RA", ra_service_name, None, f"code={response.status_code}")

    try:
        attestation_status = False

        ra_res = json.loads(response.text).get("Result")
        if not ra_res:
            logger.error(
                f"attest_client failed, verify_result is NULL, ra interface response={response.text}"
            )
            return False

        for cluster_info, res_info in ra_res.items():
            raw_evidence = res_info.get("evidence")
            tmp = json.loads(raw_evidence)
            raw_evidence_json = json.dumps(tmp, indent=4)
            logger.debug(f"get ra evidence:{raw_evidence_json}")

            token_dict = {}
            raw_token = res_info.get("token")
            token_parts = raw_token.split(".")
            if len(token_parts) >= 2:
                token_data = token_parts[1]
                try:
                    # 使用base64.urlsafe_b64decode进行URL安全模式的解码
                    decoded_data = base64.urlsafe_b64decode(
                        token_data + "=" * (4 - len(token_data) % 4)
                    ).decode("utf-8")

                    token_dict = json.loads(decoded_data)
                except Exception as e:
                    logger.exception(f"解码出现错误: {e}")
                    raise error.ServiceError(
                        "RA", ra_service_name, None, f"code={response.status_code}"
                    )

            # 取第一个节点验证就可以
            attestation_status = len(token_dict.get("policies_matched")) > 0
            # public_key_info = res_info.get("key_info").get("pub_key_info")
            break

        return attestation_status
    except Exception as e:
        logger.exception(f"Response is not JSON: client_service_name={ra_service_name}")
        raise error.ServiceError("RA", ra_service_name, None, "not JSON") from e
